RevivalTracker = LibStub("AceAddon-3.0"):NewAddon("RevivalTracker", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")

local Media = LibStub("LibSharedMedia-3.0")
local AGSMW = LibStub("AceGUISharedMediaWidgets-1.0")

local options = {
	name = "RevivalTracker",
	type = "group",
	args = {
		general = {
			name = "General",
			type = "group",
			order = 1,
			args = {
				locked = {
					name = "Locked",
					desc = "Lock the frame",
					type = "toggle",
					width = "full",
					order = 1,
					set = 
					function(info,val) 
						RevivalTracker.db.profile.locked = val 
						if val then
							RevivalTracker:Lock()
						else
							RevivalTracker:Unlock()
						end
					end,
					get = function(info) return RevivalTracker.db.profile.locked end
				},
				scale = {
					name = "Scale",
					desc = "Set the addon's scale",
					min = 0.1,
					max = 5,
					type = "range",
					width = "normal",
					order = 2,
					set = 
					function(info,val) 
						RevivalTracker.db.profile.scale = val 
						RevivalTracker.frame:SetScale(val)
					end,
					get = function(info) return RevivalTracker.db.profile.scale end
				},
				font = {
					type = "select",
					dialogControl = "LSM30_Font",
					name = "Font",
					desc = "Addon Font",
					width = "normal",
					order = 3,
					values = Media:HashTable("font"),
					get = function()
						return RevivalTracker.db.profile.font
					end,
					set = function(self,key)
						RevivalTracker.db.profile.font = key
						RevivalTracker:SetFont()
					end
				},
				headerBg = {
					name = "Background",
					type = "header",
					order = 4
				},
				background = {
					type = "select",
					dialogControl = "LSM30_Background",
					name = "Background",
					desc = "Background Texture",
					width = "normal",
					order = 5,
					values = Media:HashTable("background"),
					get = function()
						return RevivalTracker.db.profile.background
					end,
					set = function(self,key)
						RevivalTracker.db.profile.background = key
						RevivalTracker:SetBackground()
					end
				},
				backgroundColor = {
					type = "color",
					hasAlpha = true,
					name = "Background Color",
					desc = "Background Color",
					width = "normal",
					order = 6,
					get = function()
						local bgColor = RevivalTracker.db.profile.backgroundColor
						return bgColor.r, bgColor.g, bgColor.b, bgColor.a
					end,
					set = function(self,r,g,b,a)
						RevivalTracker.db.profile.backgroundColor = { r = r, g = g, b = b, a = a }
						RevivalTracker:SetBackground()
					end
				},
				headerBorder = {
					name = "Border",
					type = "header",
					order = 7
				},
				border = {
					type = "select",
					dialogControl = "LSM30_Border",
					name = "Border",
					desc = "Border Texture",
					width = "normal",
					order = 8,
					values = Media:HashTable("border"),
					get = function()
						return RevivalTracker.db.profile.border
					end,
					set = function(self,key)
						RevivalTracker.db.profile.border = key
						RevivalTracker:SetBackground()
					end
				},
				borderColor = {
					type = "color",
					hasAlpha = true,
					name = "Border Color",
					desc = "Border Color",
					width = "normal",
					order = 9,
					get = function()
						local borderColor = RevivalTracker.db.profile.borderColor
						return borderColor.r, borderColor.g, borderColor.b, borderColor.a
					end,
					set = function(self,r,g,b,a)
						RevivalTracker.db.profile.borderColor = { r = r, g = g, b = b, a = a }
						RevivalTracker:SetBackground()
					end
				},
				headerEstimation = {
					name = "Estimation",
					type = "header",
					order = 10
				},
				useMulti = {
					name = "Use multistrike",
					desc = "Use multistrike in healing estimations",
					type = "toggle",
					width = "normal",
					order = 11,
					set = function(info,val) RevivalTracker.db.profile.useMulti = val end,
					get = function(info) return RevivalTracker.db.profile.useMulti end
				},
				useCrit = {
					name = "Use crit",
					desc = "use crit in healing estimations",
					type = "toggle",
					width = "normal",
					order = 12,
					set = function(info,val) RevivalTracker.db.profile.useCrit = val end,
					get = function(info) return RevivalTracker.db.profile.useCrit end
				}
			}
		}
	}
}

local defaults = {
	profile = {
		scale = 1.0,
		revivalSpellID = 115310,
		revivalDescPattern = "Heals all party and raid members within .- yards for (.-) and clears them of all harmful Magical, Poison, and Disease effects.",
		font = "Friz Quadrata TT",
		background = "Solid",
		backgroundColor = { r = 0, g = 0, b = 0, a = 1 },
		border = "Blizzard Tooltip",
		borderColor = { r = 0, g = 0, b = 0, a = 1 },
		height = 96,
		width = 96,
		locked = false,
		effiencyCutoffs = { red = 30, white = 70 },
		idleCutoffs = { green = 30, gray = 60 },
		useCrit = false,
		useMulti = false
	}
}

function RevivalTracker:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("RevivalTrackerDB", defaults, true)
	
	LibStub("AceConfig-3.0"):RegisterOptionsTable("RevivalTracker", options);
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db);

	local acd = LibStub("AceConfigDialog-3.0")
	self.optionsFrame = acd:AddToBlizOptions("RevivalTracker", "Revival Tracker", nil, "general");
	self.optionsFrame.profile = acd:AddToBlizOptions("RevivalTracker", "Profiles", "Revival Tracker", "profile");
	
	self:RegisterChatCommand("revt", "ChatCommand")
    self:RegisterChatCommand("revivaltracker", "ChatCommand")
	
	self:CreateFrame()
	
	self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
	self:RegisterEvent("PLAYER_LOGIN")
end

function RevivalTracker:OnEnable()
	RevivalTracker:Print("Revival Tracker enabled!")
	RevivalTracker:Print("Using profile: "..self.db:GetCurrentProfile())
end

function RevivalTracker:OnDisable()
	RevivalTracker:Print("Revival Tracker disabled!")
end

function RevivalTracker:ChatCommand(input)
    if not input or input:trim() == "" then
        InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
    else
        LibStub("AceConfigCmd-3.0"):HandleCommand("revivaltracker", "RevivalTracker", input)
    end
end

function RevivalTracker:CreateFrame()
	local frame = CreateFrame("Frame", "RevivalTrackerFrame", UIParent)
	self.frame = frame
	frame:SetPoint("CENTER", UIParent )
	frame:SetSize(self.db.profile.width, self.db.profile.height)
	frame:SetScale(self.db.profile.scale)
	frame:SetMinResize(96, 96)
	frame:SetMovable(true);
	frame:EnableMouse(true);
	frame:SetUserPlaced(true);
	frame:SetClampedToScreen( true )

	self:SetBackground()
	
	if self.db.profile.locked then
		self:Lock()
	else
		self:Unlock()
	end
	
	local textCenter = frame:CreateFontString(nil, "OVERLAY")
	textCenter:SetPoint("CENTER", frame, "CENTER", 0, 0)
	textCenter:SetJustifyH("CENTER")
	textCenter:SetShadowOffset(1, -1)
	textCenter:SetTextColor(1, 1, 1)
	
	local textTop = frame:CreateFontString(nil, "OVERLAY")
	textTop:SetPoint("TOP", frame, "TOP", 0, -10)
	textTop:SetJustifyH("TOP")
	textTop:SetShadowOffset(1, -1)
	textTop:SetTextColor(0, 1, 0)	
	
	local textBottom = frame:CreateFontString(nil, "OVERLAY")
	textBottom:SetPoint("BOTTOM", frame, "BOTTOM", 0, 10)
	textBottom:SetJustifyH("BOTTOM")
	textBottom:SetShadowOffset(1, -1)
	textBottom:SetTextColor(0, 1, 0)
	
	frame.textCenter = textCenter
	frame.textTop = textTop
	frame.textBottom = textBottom
	
	self:SetFont()
	
	textCenter:SetText("")
	textTop:SetText("")
	textBottom:SetText("")
end

function RevivalTracker:SetFont()
	local font = Media:Fetch("font", self.db.profile.font)
	
	self.frame.textCenter:SetFont(font, 24, "OUTLINE")
	self.frame.textTop:SetFont(font, 10, "OUTLINE")
	self.frame.textBottom:SetFont(font, 10, "OUTLINE")
end

function RevivalTracker:SetBackground()
	local bgFile = Media:Fetch("background", self.db.profile.background)
	local edgeFile = Media:Fetch("border", self.db.profile.border)

	self.frame:SetBackdrop({
		bgFile = bgFile,
		edgeFile = edgeFile,
		tile = false, 
		edgeSize = 16, 
        insets = { left = 4, right = 4, top = 4, bottom = 4 }
	})
	
	local bgColor = self.db.profile.backgroundColor
	self.frame:SetBackdropColor(bgColor.r, bgColor.g, bgColor.b, bgColor.a)
	
	local borderColor = self.db.profile.borderColor
	self.frame:SetBackdropBorderColor(borderColor.r, borderColor.g, borderColor.b, borderColor.a)
end

function RevivalTracker:Unlock()
	local frame = self.frame
	frame:RegisterForDrag("LeftButton")
	frame:SetScript("OnDragStart", 
		function(self, button) 
			if button == "LeftButton" then
				self:StartMoving()
				self.isMoving = true
				self.hasMoved = false
			end
		end
	)
	frame:SetScript("OnDragStop", frame.StopMovingOrSizing)
end

function RevivalTracker:Lock()
	local frame = self.frame
	frame:RegisterForDrag(nil)
	frame:SetScript("OnDragStart", nil)
	frame:SetScript("OnDragStop", nil)
end

function RevivalTracker:CheckSpecialization()
	self.enabled = true

	local class, classFileName = UnitClass("player")
	if classFileName ~= "MONK" then
		self.enabled = false
	end
	
	local spec = GetSpecialization()
	if spec ~= 2 then
		self.enabled = false
	end
		
	if self.enabled then
		self:UnregisterEvents()
		self:RegisterEvents()
		self.frame:Show()
	else
		self.frame:Hide()
		self:UnregisterEvents()
	end
end

function RevivalTracker:RegisterEvents()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "Update")
	self:RegisterEvent("COMBAT_RATING_UPDATE", "Update")
	self:RegisterEvent("UNIT_SPELL_HASTE", "Update")
	self:RegisterEvent("UNIT_STATS", "Update")
	self:RegisterEvent("SPELL_POWER_CHANGED", "Update")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	
	self.cdTimer = self:ScheduleRepeatingTimer("TimerCooldown", 0.2)
end

function RevivalTracker:UnregisterEvents()
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:UnregisterEvent("COMBAT_RATING_UPDATE")
	self:UnregisterEvent("UNIT_SPELL_HASTE")
	self:UnregisterEvent("UNIT_STATS")
	self:UnregisterEvent("SPELL_POWER_CHANGED")
	self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	
	if self.cdTimer then
		self:CancelTimer(self.cdTimer)
	end
end

function RevivalTracker:UNIT_SPELLCAST_SUCCEEDED(name, unitID, spell, rank, lineID, spellID)
	if unitID ~= "player" then
		return
	end
	
	if spellID ~= self.db.profile.revivalSpellID then
		return
	end
	
	self.lastCast = GetTime()
end

function RevivalTracker:TimerCooldown()
	self:UpdateCooldown()
end

function RevivalTracker:PLAYER_REGEN_DISABLED(name, ...)
	local start, duration, enable = GetSpellCooldown(self.db.profile.revivalSpellID)

	if start == 0 or duration == 0 then
		self.lastCast = GetTime()
	end
end

function RevivalTracker:PLAYER_REGEN_ENABLED(name, ...)
	self.lastCast = nil
end

function RevivalTracker:UpdateCooldown()
	local start, duration, enable = GetSpellCooldown(self.db.profile.revivalSpellID)
	if (start > 0 and duration > 1.5) then
		self.onCooldown = true
		self.frame.textTop:SetTextColor(0.75, 0.75, 0.75)
		self.frame.textTop:SetText(SecondsToTime(start + duration - GetTime()))
	else
		if self.onCooldown then
			self.onCooldown = false
			self.lastCast = GetTime()
		end
		
		if not (UnitAffectingCombat("player") or InCombatLockdown()) then
			self.lastCast = nil
		end
	
		if self.lastCast then
			local timeSince = GetTime() - self.lastCast
			if timeSince < self.db.profile.idleCutoffs.green then
				self.frame.textTop:SetTextColor(0, 1, 0)
			elseif timeSince < self.db.profile.idleCutoffs.gray then
				self.frame.textTop:SetTextColor(0.75, 0.75, 0.75)
			else
				self.frame.textTop:SetTextColor(1, 0, 0)
			end
			
			self.frame.textTop:SetText("+" .. SecondsToTime(timeSince))
		else
			self.frame.textTop:SetTextColor(0, 1, 0)
			self.frame.textTop:SetText("Ready")
		end		
	end
	
	self:Update()
end

function RevivalTracker:ACTIVE_TALENT_GROUP_CHANGED(name, ...)
	self:CheckSpecialization()
	self:Update()
end

function RevivalTracker:PLAYER_LOGIN(name, ...)
	self:CheckSpecialization()
	self:Update()
end

function RevivalTracker:Update()
	if not self.enabled then
		return
	end
	
	local desc = GetSpellDescription(self.db.profile.revivalSpellID)
	local revivalAmount = tonumber((desc:match(self.db.profile.revivalDescPattern):gsub(",","")))
	
	if self.db.profile.useCrit then
		local critFactor = (GetSpellCritChance(4)/100) + 1
		revivalAmount = revivalAmount * critFactor
	end
	
	if self.db.profile.useMulti then
		local multiFactor = ((GetMultistrike()/100) * 2 * 0.3) + 1
		revivalAmount = revivalAmount * multiFactor
	end
	
	local members = GetNumGroupMembers()
	local groupType = "party"
	if IsInRaid() then
		groupType = "raid"
	end
	
	local totalMembers = math.max(1, members)
	local maxHeal = totalMembers * revivalAmount
	local estimatedHeal = 0
	
	local unit = "player", delta, inRange
	if members and members > 0 then
		for i = 1, members, 1 do
			unit = groupType .. i
			
			inRange = IsSpellInRange("Revival", unit)
			if inRange then
				delta = UnitHealthMax(unit) - UnitHealth(unit)
				estimatedHeal = estimatedHeal + math.min(delta, revivalAmount)
			else
				totalMembers = totalMembers - 1
			end		
		end
	else
		delta = UnitHealthMax(unit) - UnitHealth(unit)
		estimatedHeal = math.min(delta, revivalAmount)
	end

	local pct = (estimatedHeal / maxHeal) * 100
	local textPct = string.format("%.0f%%", pct)
	if pct < self.db.profile.effiencyCutoffs.red then
		self.frame.textCenter:SetTextColor(1, 0, 0)
	elseif pct < self.db.profile.effiencyCutoffs.white then
		self.frame.textCenter:SetTextColor(1, 1, 1)
	else
		self.frame.textCenter:SetTextColor(0, 1, 0)
	end
	self.frame.textCenter:SetText(textPct)
	
	local textBottom = string.format("Est. %s (%s)", RevivalTracker:ReadableNumber(estimatedHeal), tostring(totalMembers))
	self.frame.textBottom:SetText(textBottom)
end

function RevivalTracker:ReadableNumber(num, places)
    local ret
    local placeValue = ("%%.%df"):format(places or 0)
    if not num then
        return 0
    elseif num >= 1000000000000 then
        ret = placeValue:format(num / 1000000000000) .. " Tril" -- trillion
    elseif num >= 1000000000 then
        ret = placeValue:format(num / 1000000000) .. " Bil" -- billion
    elseif num >= 1000000 then
        ret = placeValue:format(num / 1000000) .. " Mil" -- million
    elseif num >= 1000 then
        ret = placeValue:format(num / 1000) .. "k" -- thousand
    else
        ret = num -- hundreds
    end
    return ret
end